import numpy as np
from numpy.fft import rfft, irfft

from roguewave.timeseries_analysis.filtering import sos_filter
from scipy.signal import butter
from scipy.signal.windows import tukey
from roguewave.wavephysics.fluidproperties import GRAVITATIONAL_ACCELERATION
import pandas as pd


def frequency_scale(depth, **kwargs):
    """
    Calculate the frequency scale for the response function. This frequency scale is simply derived from the shallow
    water relation - noting that

    $$ kd=1  $$

    corresponds to

    $$ f = \sqrt{g/d} / 2\pi $$

    if we use the shallow water dispersion relation

    $$ \omega^2 = gd k^2 $$

    :param depth: depth in meters
    :return: frequency scale in Hz
    """
    g = kwargs.get("g", GRAVITATIONAL_ACCELERATION)
    return np.sqrt(depth / g) / (2 * np.pi)


def sample_irregular_signal(time_seconds: np.ndarray, signal: np.ndarray, sampling_frequency: float):
    """
    Sample an irregular signal at a constant rate. We do this through interpolation.

    :param time_seconds: time in seconds
    :param signal: signal
    :param sampling_frequency: sampling frequency in Hz
    :return: interpolated time, interpolated signal
    """

    # Interpolate the time base to integer values
    time_base = (time_seconds - time_seconds[0]) * sampling_frequency
    interpolated_time_base = np.arange(np.floor(time_base[-1]) + 1)
    interpolated_signal = np.interp(interpolated_time_base, time_base, signal)
    return interpolated_time_base / sampling_frequency + time_seconds[0], interpolated_signal


def spectral_derivative(signal, sampling_frequency, sos=None):
    """
    Calculate the time derivative of a signal using the Fourier transform. The signal is assumed to be sampled at a
    constant rate given by the sampling frequency.

    There will be edge effects unless the signal has been windowed.

    :param signal: signal
    :param sampling_frequency: sampling frequency in Hz
    :return: time derivative of the signal
    """
    if sos is not None:
        signal = sos_filter(signal, 'filtfilt', sos=sos)

    n = len(signal)
    omega = 2 * np.pi * np.fft.rfftfreq(n, 1 / sampling_frequency)

    return irfft(1j * omega * rfft(signal, n), n)


def _non_linear_contribution(pressure_head_meter, sampling_frequency, sensor_height, **kwargs):
    """
    Calculate the non-linear contribution to the pressure signal. The non-linear contribution has the form

    $$ \\frac{1}{2g} \\frac{d}{dt} \\left( h \\frac{d h}{dt} \\right) $$

    where $h$ is the pressure head in meters. The signal is assumed to be sampled at a constant rate given by the
    sampling frequency.

    :param pressure_head_meter: Water pressure head in meter. The signal is assumed to be sampled at a constant rate
        given by the sampling frequency.
    :param sampling_frequency: Sampling frequency in Hz.
    :param sensor_height: Height (in meters) of the sensor above the bed level.

    :return: Non-linear contribution to the surface elevation in meters.
    """
    _filter = kwargs.get("filter", True)
    g = kwargs.get("gravitational_acceleration", GRAVITATIONAL_ACCELERATION)
    mean_depth = np.nanmean(pressure_head_meter) + sensor_height

    if _filter:
        relative_cutoff_frequency = kwargs.get("relative_cutoff_frequency", 2)
        cutoff_frequency = frequency_scale(mean_depth, **kwargs) * relative_cutoff_frequency

        if cutoff_frequency < sampling_frequency / 2:
            sos = kwargs.get(
                "sos",
                butter(4, cutoff_frequency, btype='lowpass', fs=sampling_frequency, output='sos')
            )
            pressure_head_meter = sos_filter(pressure_head_meter, 'filtfilt', sos=sos)

    hydrostatic_estimate = pressure_head_meter - np.nanmean(pressure_head_meter)
    dpdt = spectral_derivative(hydrostatic_estimate, sampling_frequency)
    return spectral_derivative(hydrostatic_estimate * dpdt, sampling_frequency) / 2 / g


def slope_term(pressure_in_meters, sampling_frequency, sensor_height=0,**kwargs):
    slope = kwargs.get("slope", 0)
    n = len(pressure_in_meters)

    mean_depth = np.nanmean(pressure_in_meters) + sensor_height
    omega = 2 * np.pi * np.fft.rfftfreq(n, 1 / sampling_frequency)
    g = kwargs.get("gravitational_acceleration", GRAVITATIONAL_ACCELERATION)
    amplification_factor = 1j * omega * slope * np.sqrt( g / mean_depth)/2

    return irfft(rfft(pressure_in_meters, n) * amplification_factor, n)


def non_hydrostatic_contribution(pressure_in_meters, sampling_frequency, sensor_height=0, **kwargs):
    """
    Calculate the non-hydrostatic contribution to the pressure signal. Here we use a truncated Taylor series expansion
    of the full expression.

    :param pressure_in_meters: Water pressure in meters. The signal is assumed to be sampled at a constant rate given by
        the sampling frequency.
    :param sampling_frequency: Sampling frequency in Hz.
    :param sensor_height: Height (in meters) of the sensor above the bed level.

    :return: Non-hydrostatic contribution to the surface elevation in meters.
    """
    _filter = kwargs.get("filter", True)
    g = kwargs.get("gravitational_acceleration", GRAVITATIONAL_ACCELERATION)
    current = kwargs.get('current_ms', 0)
    current_angle = kwargs.get('wave_current_mutual_angle_rad', 0)
    mean_depth = np.nanmean(pressure_in_meters) + sensor_height

    if _filter:
        relative_cutoff_frequency = kwargs.get("relative_cutoff_frequency", 2)
        cutoff_frequency = frequency_scale(mean_depth, **kwargs) * relative_cutoff_frequency

        if cutoff_frequency < sampling_frequency / 2:
            sos = kwargs.get(
                "sos",
                butter(4, cutoff_frequency, btype='lowpass', fs=sampling_frequency, output='sos')
            )
            pressure_in_meters = sos_filter(pressure_in_meters, 'filtfilt', sos=sos)

    order = kwargs.get("order", 4)
    taper = kwargs.get("taper", True)

    n = len(pressure_in_meters)
    omega = 2 * np.pi * np.fft.rfftfreq(n, 1 / sampling_frequency)
    depth = np.nanmean(pressure_in_meters) + sensor_height

    relative_omega = omega * np.sqrt(depth / g) * np.sqrt((depth ** 2 - sensor_height ** 2) / (depth ** 2))
    if taper:
        relative_omega = np.sqrt(5 * np.tanh(relative_omega ** 2 / 5))

    relative_current = np.cos(current_angle)*current / np.sqrt(g * depth)

    fac = np.ones(4)
    fac[order:] = 0
    amplification_factor = (
            fac[1] * relative_omega ** 2 / 2 +
            fac[2] * 5 * relative_omega ** 4 / 24 +
            fac[3] * 37 * relative_omega ** 6 / 720 -
            relative_current * relative_omega ** 2
    )

    return irfft(rfft(pressure_in_meters, n) * amplification_factor, n)


def surface_elevation_from_pressure(
        pressure_head_meter,
        sampling_frequency,
        sensor_height=0,
        **kwargs
) -> pd.DataFrame:
    """
    Calculate the surface elevation from a pressure head signal (pressure divided by water density and gravitational
    acceleration). The signal is assumed to be sampled at a constant rate given by the sampling frequency.

    :param pressure_head_meter: Water pressure head in meter. The signal is assumed to be sampled at a constant rate
        given by the sampling frequency.
    :param sampling_frequency: Sampling frequency in Hz.
    :param sensor_height: Height (in meters) of the sensor above the bed level.
    :return: Zero-mean Surface elevation in meters.
    """

    # Window
    # ------
    # Because we use Fourier transforms to calculate the non-hydrostatic/nonlinear contributions, we need to window the
    # signal to avoid edge effects. The default window is a Tukey window that tapers the nonhydrostatic/nonlinear signal
    # to zero at the edges.
    number_of_taper_points = kwargs.get("number_of_taper_points", 20)
    taper_fraction_in_default_window = number_of_taper_points / len(pressure_head_meter)
    window = kwargs.get("window", tukey(len(pressure_head_meter),taper_fraction_in_default_window))

    # Calculate mean water levels
    # ---------------------------
    mean_pressure_head = np.nanmean(pressure_head_meter)
    mean_depth = mean_pressure_head + sensor_height

    # Hydrostatic estimate
    # --------------------
    dataframe = pd.DataFrame()
    dataframe['surface elevation (meter)'] = pressure_head_meter - mean_pressure_head
    dataframe['depth (meter)'] = mean_depth

    # For very short signals we cannot do Fourier analysis- return the hydrostatic estimate
    if len(pressure_head_meter) < 100:
        return dataframe

    # Non-Hydrostatic estimate
    # ------------------------
    dataframe['surface elevation (meter)'] += window * non_hydrostatic_contribution(
        pressure_head_meter, sampling_frequency, sensor_height, **kwargs)

    # Slope estimate
    # ------------------------
    dataframe['surface elevation (meter)'] += window * slope_term(
        pressure_head_meter, sampling_frequency, sensor_height, **kwargs)

    # Nonlinear estimate
    # ------------------
    nonlinear_sign = kwargs.get("non_linear_sign", 1)
    if kwargs.get("nonlinear_contribution", False):
        dataframe['surface elevation (meter)'] += nonlinear_sign*(
                _non_linear_contribution(pressure_head_meter, sampling_frequency, sensor_height) * window)

    return dataframe



import numpy
from roguewave.tools.math import wrapped_difference
from roguewave.tools.grid import enclosing_points_1d


def interpolate_periodic(
    xp: numpy.ndarray,
    fp: numpy.ndarray,
    x: numpy.ndarray,
    x_period: float = None,
    fp_period: float = None,
    fp_discont: float = None,
    left: float = numpy.nan,
    right: float = numpy.nan,
) -> numpy.ndarray:
    """
    Interpolation function that works with periodic domains and periodic ranges
    :param xp:
    :param fp:
    :param x:
    :param x_period:
    :param fp_period:
    :param fp_discont:
    :return:
    """

    indices = enclosing_points_1d(xp, x, period=x_period, regular_xp=False)

    delta_x = wrapped_difference(x - xp[indices[0, :]], period=x_period)
    delta_xp = wrapped_difference(
        xp[indices[1, :]] - xp[indices[0, :]], period=x_period
    )
    delta_fp = wrapped_difference(
        delta=fp[indices[1, :]] - fp[indices[0, :]], period=fp_period
    )
    fp0 = fp[indices[0, :]]

    if x_period is not None:
        interpolation = fp0 + delta_fp * delta_x / delta_xp

    else:
        # constant extrapolation
        mask = delta_xp.astype("float64") > 0.0
        interpolation = numpy.zeros((len(x),))
        interpolation[mask] = (
            fp0[mask] + delta_fp[mask] * delta_x[mask] / delta_xp[mask]
        )
        interpolation[~mask] = fp0[~mask]

        # substitute value if given
        if left is not None:
            interpolation = numpy.where(x < xp[0], left, interpolation)

        # substitute value if given
        if right is not None:
            interpolation = numpy.where(x > xp[-1], right, interpolation)

    mask = numpy.isfinite(interpolation)
    interpolation[mask] = wrapped_difference(
        delta=interpolation[mask], period=fp_period, discont=fp_discont
    )

    return interpolation


def interpolation_weights_1d(
    xp,
    x,
    indices,
    period: float = None,
    extrapolate_left=True,
    extrapolate_right=True,
    nearest_neighbour=False,
):
    """
    Find the weights for the linear interpolation problem given a set of
    indices such that:

                xp[indices[0,:]] <= x[:] < xp[indices[1,:]]

    Indices are assumed to be as returned by "enclosing_points_1d" (see
    roguewave.tools.grid).

    Returns weights (nx,2) to perform the one-dimensional piecewise linear
    interpolation to a function given at discrete datapoints xp and evaluated
    at x. Say that at all xp we have for the function values fp,
    the interpolation would then be

        f = fp[ Indices[1,:] ]  *  weights[1,:] +
            fp[ Indices[2,:] ]  *  weights[2,:]

    :param xp:
    :param x:
    :param indices:
    :param period:
    :return:
    """

    weights = numpy.empty_like(indices, dtype="float64")

    if xp[-1] < xp[0]:
        # make sure we are in a coordinate frame where the vector is
        # in ascending order.
        x = xp[0] - x
        xp = xp[0] - xp

    delta_x = wrapped_difference(x - xp[indices[0, :]], period=period)

    delta_xp = wrapped_difference(xp[indices[1, :]] - xp[indices[0, :]], period=period)

    if period is None:
        mask = (x >= xp[0]) & (x < xp[-1])
        frac = numpy.empty((len(x),))
        frac[mask] = delta_x[mask] / delta_xp[mask]
        if extrapolate_left:
            frac[x < xp[0]] = 1
        else:
            frac[x < xp[0]] = numpy.nan

        if extrapolate_right:
            frac[x > xp[-1]] = 0
        else:
            frac[x > xp[-1]] = numpy.nan

        # At the right end point to avoid devision by 0.
        frac[x == xp[-1]] = 0.0
    else:
        frac = delta_x / delta_xp

    if nearest_neighbour:
        frac = numpy.rint(frac)

    weights[0, :] = 1 - frac
    weights[1, :] = frac

    return weights



from numpy.fft import rfft, irfft
import numpy as np
import scipy
from scipy.signal.windows import tukey

def sample_irregular_signal(time_seconds: np.ndarray, signal: np.ndarray, sampling_frequency: float):
    """
    Sample an irregular signal at a constant rate. We do this through interpolation.

    :param time_seconds: time in seconds
    :param signal: signal
    :param sampling_frequency: sampling frequency in Hz
    :return: interpolated time, interpolated signal
    """

    # Interpolate the time base to integer values
    time_base = (time_seconds - time_seconds[0]) * sampling_frequency
    interpolated_time_base = np.arange(np.floor(time_base[-1]) + 1)
    interpolator = scipy.interpolate.interp1d( time_base, signal,kind='quadratic',assume_sorted=True)
    interpolated_signal = interpolator(interpolated_time_base)
    return interpolated_time_base / sampling_frequency + time_seconds[0], interpolated_signal

def upsample(signal, factor: int, t0=0, sampling_frequency=2.5):
    """
    Spectral upsampling. There will be edge effects unless the signal has been windowed.

    :param signal:
    :param factor:
    :return:
    """
    n = len(signal)
    upsampled_time = (
        np.linspace(0, n * factor, n * factor, endpoint=False)
        / (sampling_frequency * factor)
        + t0
    )

    return upsampled_time, irfft(rfft(signal) * factor, n * factor)

def resample( sample_time, time, signal):

    desired_sampling_frequency = 1/np.diff(sample_time).min()

    sampling_frequency = 1/np.diff(time).mean()
    relative_sampling_time = time * sampling_frequency % 1 - 0.5
    if np.abs(relative_sampling_time).max() > 0.01:
        regular_time, regular_sampled_signal = sample_irregular_signal(time, signal, sampling_frequency)
    else:
        regular_time = time
        regular_sampled_signal = signal

    relative_difference = desired_sampling_frequency / sampling_frequency

    upsample_factor = max( 5*int(np.ceil(relative_difference)), 5)
    upsampled_time, upsampled_signal_spectral \
        = upsample( regular_sampled_signal, upsample_factor, regular_time[0], sampling_frequency )

    upsampled_signal_spline = scipy.interpolate.interp1d( regular_time, regular_sampled_signal,
                                                          kind='quadratic', assume_sorted=True)(upsampled_time)

    delta = upsampled_signal_spectral - upsampled_signal_spline

    upsampled_signal = upsampled_signal_spline + tukey( len(upsampled_signal_spline), 0.01) * delta
    return np.interp( sample_time, upsampled_time, upsampled_signal, left=np.nan, right=np.nan)



import numpy as np
import matplotlib.pyplot as plt
from numba import njit

@njit(cache=True)
def lower_index_zero_interval(signal):
    """
    Define waves as the zero crossings of the signal.
    """
    return np.where(np.diff(np.sign(signal))<0)[0]

def wave_mask(signal):
    """
    Define waves as the zero crossings of the signal.
    """
    return np.cumsum(np.diff(np.sign(signal))<0)

def time_zero(time,signal):
    lower_index = lower_index_zero_interval(signal)

    if lower_index[-1] == len(signal):
        lower_index = lower_index[:-1]

    upper_index = lower_index + 1

    delta_signal = -signal[lower_index]/(signal[upper_index]-signal[lower_index])
    delta_time = time[upper_index]-time[lower_index]

    zero_crossing_time = time[lower_index] + signal[lower_index] * delta_time/ delta_signal
    return zero_crossing_time

@njit(cache=True)
def zero_crossing_extrema(signal):
    """
    Find the extrema of a signal.
    """
    indices = lower_index_zero_interval(signal)

    number_of_waves = len(indices) - 1
    maxima = np.zeros(number_of_waves) - np.inf
    minima = np.zeros(number_of_waves) + np.inf

    for wave_index in range(0,number_of_waves):
        lower_index = indices[wave_index]
        upper_index = indices[wave_index+1]

        for signal_index in range(lower_index,upper_index):
            if signal[signal_index] > maxima[wave_index]:
                maxima[wave_index] = signal[signal_index]
            if signal[signal_index] < minima[wave_index]:
                minima[wave_index] = signal[signal_index]

    return maxima, minima

def zero_crossing_wave_heights(signal):
    maxima, minima = zero_crossing_extrema(signal)
    return maxima - minima

def zero_crossing_period(time, signal):
    zero_crossing_time = time_zero(time,signal)
    return np.diff(zero_crossing_time)

def maximum_wave_height(signal):
    _waveheights = zero_crossing_wave_heights(signal)
    return np.max(_waveheights)




import typing

import numpy as np
import pandas
import pandas as pd

from roguewave.timeseries_analysis import (
    pipeline,
    DEFAULT_SPOTTER_PIPELINE,
    DEFAULT_DISPLACEMENT_PIPELINE,
)
from roguewave.tools.time_integration import (
    cumulative_distance,
)
from roguewave.timeseries_analysis import estimate_frequency_spectrum
from roguewave.tools.time import datetime64_to_timestamp
from pandas import DataFrame, concat
from roguewavespectrum import Spectrum
from .read_csv_data import read_displacement, read_gps, read_rbr, read_baro
from .parser import apply_to_group
from roguewavespectrum.spotter._spotter_post_processing import spotter_frequency_response_correction
from roguewave.spotter._pressure_analysis import surface_elevation_from_pressure, sample_irregular_signal, frequency_scale
from roguewave.wavephysics.fluidproperties import WATER_DENSITY, GRAVITATIONAL_ACCELERATION
from scipy.signal import butter
import os
from roguewave.timeseries_analysis.filtering import sos_filter
from roguewave.timeseries_analysis.sampling import resample
from roguewave.spotter.parser import _netcdf_filename, save_as_netcdf
import xarray as xr

LAST_BIN_WIDTH = 0.3
LAST_BIN_FREQUENCY_START = 0.5
LAST_BIN_FREQUENCY_END = 0.8
SPOTTER_FREQUENCY_RESOLUTION = 2.5 / 256


def displacement_from_gps_doppler_velocities(
        path, pipeline_stages=None, cache_as_netcdf=False, **kwargs
) -> DataFrame:
    if pipeline_stages is None:
        pipeline_stages = DEFAULT_SPOTTER_PIPELINE

    doppler_velocities = read_gps(path, cache_as_netcdf=cache_as_netcdf)

    def process(data: DataFrame):
        time = data["time"].values
        x = pipeline(time, data["u"].values, pipeline_stages)
        y = pipeline(time, data["v"].values, pipeline_stages)
        z = pipeline(time, data["w"].values, pipeline_stages)

        return DataFrame(
            data={"time": time, "x": x, "y": y, "z": z, "group id": data["group id"]}
        )

    return apply_to_group(process, doppler_velocities)


def displacement_from_gps_positions(path) -> DataFrame:
    pipeline_stages = DEFAULT_DISPLACEMENT_PIPELINE

    gps_location_data = read_gps(path, postprocess=True)

    def get_cumulative_distances(data: DataFrame):
        time = data["time"].values
        x, y = cumulative_distance(data["latitude"].values, data["longitude"].values)
        z = data["z"].values

        return DataFrame(
            data={"time": time, "x": x, "y": y, "z": z, "group id": data["group id"]}
        )

    cum_distance = apply_to_group(get_cumulative_distances, gps_location_data)

    def process(data: DataFrame):
        time = data["time"].values
        x = pipeline(time, data["x"].values, pipeline_stages)
        y = pipeline(time, data["y"].values, pipeline_stages)
        z = pipeline(time, data["z"].values, pipeline_stages)

        return DataFrame(
            data={"time": time, "x": x, "y": y, "z": z, "group id": data["group id"]}
        )

    return apply_to_group(process, cum_distance)


def spectra_from_raw_gps(
        path=None, displacement_doppler=None, displacement_location=None, **kwargs
) -> Spectrum:
    if displacement_doppler is None:
        displacement_doppler = displacement_from_gps_doppler_velocities(path, **kwargs)

    if displacement_location is None:
        try:
            displacement_location = displacement_from_gps_positions(path)
        except:
            displacement_location = None

    correct_for_numerical_integration = kwargs.get("response_correction", True)
    order = kwargs.get("order", 4)
    n = kwargs.get("n", 1)

    if kwargs.get("use_u", False):
        x = displacement_doppler["x"].values
    else:
        x = displacement_location["x"].values

    if kwargs.get("use_v", False):
        y = displacement_doppler["y"].values
    else:
        y = displacement_location["y"].values

    if kwargs.get("use_w", True):
        z = displacement_doppler["z"].values
    else:
        z = displacement_location["z"].values

    time = datetime64_to_timestamp(displacement_doppler["time"].values)
    spectrum = estimate_frequency_spectrum(time, x, y, z, **kwargs)

    if correct_for_numerical_integration:
        spectrum = spotter_frequency_response_correction(spectrum, order, n)
    return spectrum


def spectra_from_displacement(path, **kwargs) -> Spectrum:
    displacement = read_displacement(path,cache_as_netcdf=True)
    time = datetime64_to_timestamp(displacement["time"].values)

    kwargs = kwargs.copy()
    correct_for_numerical_integration = kwargs.pop("response_correction", True)
    order = kwargs.pop("order", 4)
    n = kwargs.pop("n", 1)

    spectrum = estimate_frequency_spectrum(
        time,
        displacement["x"].values,
        displacement["y"].values,
        displacement["z"].values,
        **kwargs
    )
    if correct_for_numerical_integration:
        spectrum = spotter_frequency_response_correction(spectrum, order, n)
    return spectrum


def surface_elevation_from_rbr(
        path,
        sensor_type,
        sensor_height=0,
        sampling_frequency=2,
        cache_as_netcdf=True,
        **kwargs):

    data = read_rbr(path, sensor_type=sensor_type, cache_as_netcdf=cache_as_netcdf, **kwargs)
    water_density = kwargs.get("density", WATER_DENSITY)
    gravitational_acceleration = kwargs.get("gravitational_acceleration", GRAVITATIONAL_ACCELERATION)

    MINIMUM_SAMPLE_LENGTH = 900


    dataframes = []
    for group in data['group id'].unique():
        data_group = data[ data['group id'] == group]
        pressure_head_meter = data_group['pressure (pascal)'].values / water_density / gravitational_acceleration
        irregular_time_in_seconds =data_group['time'].values.astype(float)/1e9

        # If the pressure head is not a finite number/nan - exclude
        if not np.all(np.isfinite(pressure_head_meter)):
            continue

        # If depth is negative - exclude
        if np.nanmean(data_group['pressure (pascal)'].values) < 0:
            continue

        # If the sample is too short - exclude
        if len(irregular_time_in_seconds) < MINIMUM_SAMPLE_LENGTH:
            continue

        time_in_seconds,p = sample_irregular_signal(
            irregular_time_in_seconds,
            pressure_head_meter,
            sampling_frequency
        )

        dataframe = surface_elevation_from_pressure(p,sampling_frequency,sensor_height=sensor_height,**kwargs)

        # If the surface elevation is ever higher than the depth - exclude
        if dataframe['surface elevation (meter)'].max() > dataframe['depth (meter)'].max():
            continue

        # If the surface elevation is ever higher lower than the depth - exclude
        if dataframe['surface elevation (meter)'].min() < -dataframe['depth (meter)'].max():
            continue

        dataframe['time'] = pandas.to_datetime(time_in_seconds,unit='s')
        dataframe['group id'] = group
        dataframes.append(dataframe)

    return pd.concat(dataframes)


def surface_elevation_from_rbr_and_spotter(
        path,
        sensor_type,
        sensor_height=0,
        sampling_frequency_rbr=2,
        bandpass=True,
        cache_as_netcdf=True,
        **kwargs
):
    _str = f'_{sensor_type}_{sensor_height}_{sampling_frequency_rbr}_{bandpass}'
    for key in kwargs:
        _str += f'_{key}={kwargs[key]}'

    filename = f'{sensor_type}_and_spotter'+_str
    filepath = _netcdf_filename(path,filename  )
    if cache_as_netcdf and os.path.exists(filepath):
        return xr.open_dataset(filepath).to_dataframe()

    rbr = surface_elevation_from_rbr(path, sensor_type, sensor_height,sampling_frequency=sampling_frequency_rbr,
                                     cache_as_netcdf=cache_as_netcdf, **kwargs)

    spotter = read_displacement(path, cache_as_netcdf=cache_as_netcdf, postprocess=True)

    data = DataFrame()


    spotter_values_z = spotter['z'].values
    spotter_values_x = spotter['x'].values
    spotter_values_y = spotter['y'].values
    spotter_time = spotter['time'].values.astype(float) / 1e9

    rbr_time = rbr['time'].values.astype(float) / 1e9

    data['time'] = rbr['time']
    data['rbr surface elevation (meter)'] = rbr['surface elevation (meter)']
    data['spotter surface elevation (meter)'] = np.interp(rbr_time, spotter_time, spotter_values_z)
    data['spotter eastward displacement (meter)'] = np.interp(rbr_time, spotter_time, spotter_values_x)
    data['spotter northward displacement (meter)'] = np.interp(rbr_time, spotter_time, spotter_values_y)
    data['depth (meter)'] = rbr['depth (meter)']
    data['group id'] = rbr['group id']

    if bandpass:
        data = _band_pass_filter(data, sampling_frequency_rbr, **kwargs)

    if cache_as_netcdf:
        save_as_netcdf(data, path, filename)
    return data


def _band_pass_filter(data, sampling_frequency, **kwargs):
    def func(data):
        df = pd.DataFrame()
        df['time'] = data['time']


        keys = ['rbr surface elevation (meter)', 'spotter surface elevation (meter)',
                'spotter eastward displacement (meter)', 'spotter northward displacement (meter)']
        df['depth (meter)'] = data['depth (meter)']
        depth = data['depth (meter)'].values[0]
        freq = 2.*frequency_scale(depth)

        if freq < sampling_frequency/2:
            sos = butter(4, [0.033,freq], btype='bandpass', fs=sampling_frequency, output='sos')
        else:
            sos = butter(4, 0.033, btype='highpass', fs=sampling_frequency, output='sos')
        for key in keys:
            if len(data[key].values) < 100:
                df[key] = data[key]
            else:
                df[key] = sos_filter(data[key].values, 'filtfilt', sos=sos)
        return df

    return apply_to_group(func, data)


def spectra_from_rbr_and_spotter(
        path,
        sensor_type,
        window_length,
        sensor_height=0,
        sampling_frequency_rbr=2,
        bandpass=False,
        cache_as_netcdf=True,
        **kwargs) -> typing.Tuple[Spectrum,Spectrum]:

    data = surface_elevation_from_rbr_and_spotter(
        path,
        sensor_type,
        sensor_height,
        sampling_frequency_rbr,
        bandpass=bandpass,
        cache_as_netcdf=cache_as_netcdf,
        **kwargs
    )

    time = data['time'].values.astype(float) / 1e9
    rbr = data['rbr surface elevation (meter)'].values
    spot = data['spotter surface elevation (meter)'].values

    x = np.zeros_like(rbr)
    y = np.zeros_like(rbr)

    spot_spectrum = estimate_frequency_spectrum(
        time,
        data['spotter eastward displacement (meter)'].values,
        data['spotter northward displacement (meter)'].values,
        spot,
        segment_length_seconds=window_length,
        sampling_frequency=sampling_frequency_rbr,
        **kwargs
    )


    rbr_spectrum = estimate_frequency_spectrum(
        time,
        x,
        y,
        rbr,
        segment_length_seconds=window_length,
        sampling_frequency=sampling_frequency_rbr,
        **kwargs
    )


    return rbr_spectrum, spot_spectrum


import numpy as np
from scipy.signal import hilbert
import pandas as pd

def zero_mean(signal):
    return signal - np.nanmean(signal)

def third_order_moment(signal):
    z = hilbert(zero_mean(signal))
    return np.nanmean(z*z*np.conj(z))*3/4

def fourth_order_moment(signal):
    z = hilbert(zero_mean(signal))
    return 3*np.nanmean(z*z*np.conj(z)*np.conj(z))/8 + 4*np.nanmean(z*z*z*np.conj(z))/8

def normalized_third_order_moment(signal):
    moment = third_order_moment(signal)
    norm = variance(signal)**(3/2)
    return moment / norm

def skewness(signal):
    return np.real(third_order_moment(signal))

def asymmetry(signal):
    return np.imag(third_order_moment(signal))

def kurtosis(signal):
    return np.real(fourth_order_moment(signal))

def normalized_kurtosis(signal):
    return kurtosis(signal) / variance(signal)**2

def excess_kurtosis(signal):
    return normalized_kurtosis(signal) - 3

def im_kurtosis(signal):
    return np.imag(fourth_order_moment(signal))

def normalized_im_kurtosis(signal):
    return im_kurtosis(signal) / variance(signal)**2

def normalized_skewness(signal):
    return np.real(normalized_third_order_moment(signal))

def normalized_asymmetry(signal):
    return np.imag(normalized_third_order_moment(signal))

def variance(signal):
    return np.nanmean(signal**2)

def bulk_properties(signal) -> pd.DataFrame:
    df = pd.DataFrame(
        columns=[
            "variance",
            "skewness",
            "asymmetry",
            "normalized skewness",
            "normalized asymmetry",
        ],
        index=[0],
    )

    df['mean'] = np.nanmean(signal)
    df["variance"] = variance(signal)
    df["hs"] = 4*np.sqrt(variance(signal))
    df["skewness"] = skewness(signal)
    df["asymmetry"] = asymmetry(signal)
    df["normalized skewness"] = normalized_skewness(signal)
    df["normalized asymmetry"] = normalized_asymmetry(signal)
    df["kurtosis"] = kurtosis(signal)
    df["excess kurtosis"] = excess_kurtosis(signal)
    df["normalized kurtosis"] = normalized_kurtosis(signal)
    df["im kurtosis"] = im_kurtosis(signal)
    df["normalized im kurtosis"] = normalized_im_kurtosis(signal)
    return df